(* fichier difftree.ml: n-me nombre de Hamming par arbres diffrentiels *)

open Num

                 (*+-----------------------------------------+
                   |  Structure abstraite de groupe ordonn  |
                   +-----------------------------------------+*)

module type Groupe = sig
  type t                     (* type de base   *)
  val (++) : t -> t -> t     (* addition       *)
  val (--) : t -> t -> t     (* soustraction   *)
  val (<<) : t -> t -> bool  (* comparaison    *)
  val zro : t               (* constantes     *)
  val ln2  : t
  val ln3  : t
  val ln5  : t
  val string : t -> string   (* affichage *)
end

                         (*+------------------------+
                           |  Arbres diffrentiels  |
                           +------------------------+*)

module Difftree(G:Groupe) : sig

  type arbre                                        (* type abstrait            *)
  val  singleton  : G.t -> arbre                    (* arbre  un lment       *)
  val  taille     : arbre -> num                    (* taille d'un arbre        *)
  val  inc        : G.t -> arbre -> arbre           (* translation              *)
  val  dec        : G.t -> arbre -> arbre           (* translation rciproque   *)
  val  insre     : G.t -> arbre -> arbre           (* ajoute un lment        *)
  val  dcoupe    : G.t -> arbre -> arbre*arbre     (* dcoupe pr.  un pivot   *)
  val  recolle    : arbre -> arbre -> arbre         (* fusion sans recouvrement *)
  val  kme_fort : num -> num -> arbre list -> G.t (* recherche par rang       *)

end = struct
  open G

  (*                cardinal, tiquette, gauche, droite *)
  type arbre = Vide | Noeud of   num * t * arbre * arbre

  (* utilitaires *)
  let singleton a = Noeud(Int 1, a, Vide, Vide)
  and taille = function Vide -> Int 0 | Noeud(t,_,_,_) -> t
  and inc x  = function Vide -> Vide  | Noeud(t,a,g,d) -> Noeud(t,a++x,g,d)
  and dec x  = function Vide -> Vide  | Noeud(t,a,g,d) -> Noeud(t,a--x,g,d)

  (* insre un lment *)
  let rec insre x = function
    | Vide           -> singleton(x)
    | Noeud(t,a,g,d) -> if a << x then Noeud(t+/Int(1), a, g, insre (x--a) d)
                                  else Noeud(t+/Int(1), a, insre (x--a) g, d)
          
  (* dcoupe l'arbre en deux parties spars par x *)
  let rec dcoupe x = function
  | Vide           -> (Vide,Vide)
  | Noeud(t,a,g,d) -> if a << x
      then let (d1,d2) = dcoupe (x--a) d in Noeud(t-/taille(d2), a, g, d1), (inc a d2)
      else let (g1,g2) = dcoupe (x--a) g in (inc a g1), Noeud(t-/taille(g1), a, g2, d)

  (* retire le plus grand lment d'un arbre *)
  let rec maximum = function
  | Vide -> failwith "max(vide)"
  | Noeud(_,a,g,Vide) -> (a,inc a g)
  | Noeud(t,a,g,d)    -> let (x,d') = maximum(d) in (x++a),Noeud(t-/Int(1), a, g, d')

  (* recolle deux arbres sans recouvrement *)
  let recolle g d =
    if g = Vide then d
    else let (a,g') = maximum(g) in Noeud(taille(g)+/taille(d), a, dec a g', dec a d)

  (* dcoupe une fort en deux forts spares par x    *)
  (* et retourne les deux forts et leurs tailles.      *)
  (* (t1,t2,f1,f2) = rsultat partiel pour accumulation *)
  let rec dcoupe_fort x ((t1,t2,f1,f2) as accu) = function
    | []          -> accu
    | tte::queue -> let (g,d) = dcoupe x tte in
                     dcoupe_fort x (t1+/taille(g), t2+/taille(d),
                                      (if g = Vide then f1 else g::f1),
                                      (if d = Vide then f2 else d::f2)) queue

  (* extrait un arbre de taille maximale *)
  let rec max_arbre a f = function
  | []       -> (a,f)
  | b::suite -> if taille(b) >/ taille(a) then max_arbre b (a::f) suite
                                          else max_arbre a (b::f) suite

  (* extrait la k-me tiquette d'un arbre en partant de la fin *)
  let rec kme_arbre k = function
    | Vide -> failwith "arbre vide"
    | Noeud(t,a,g,d) ->
        if taille(d) =/ k then (a, inc a g, inc a d)
        else if taille(d) >/ k
        then let (x,d1,d2) = kme_arbre k d
             in (a++x, Noeud(t-/taille(d2)-/Int(1),a,g,d1), inc a d2)
        else let (x,g1,g2) = kme_arbre (k-/taille(d)-/Int(1)) g
             in (a++x, inc a g1, Noeud(t-/taille(g1)-/Int(1),a,g2,d))

  (* extrait la k-me tiquette de la fort f en partant de la fin *)
  (* t = cardinal de la fort                                      *)
  let rec kme_fort k t f = match f with
      | [] -> failwith "fort vide"
      | arbre::suite ->
          Printf.printf "k=%s t=%s\n" (string_of_num k) (string_of_num t);
          let (arbre,suite) = max_arbre arbre [] suite in
          let k' = quo_num (k*/taille(arbre)) t in
          let (a,g,d) = kme_arbre k' arbre in
          let (t1,t2,f1,f2) = dcoupe_fort a
                                (taille(g), taille(d),
                                 (if g = Vide then [] else [g]),
                                 (if d = Vide then [] else [d])) suite in
          if      t2 >/ k then kme_fort k t2 f2
          else if t2 =/ k then a
          else                 kme_fort (k-/t2-/Int(1)) t1 f1

end (* Difftree *)


                  (* +---------------------------+
                     |  n-me nombre de Hamming  |
                     +---------------------------+ *)

module Hamming(G:Groupe) : sig

  val hamming : num -> G.t

end = struct

  module D = Difftree(G)
  open G
  open D

  type ttradre = {
    mutable m2: t;         (* multiple de ln2         *)
    mutable m5: t;         (* multiple suivant de ln5 *)
    mutable n : num;       (* taille ttradre        *)
    mutable p : num;       (* taille triangle         *)
    mutable l : arbre;     (* bande d'paisseur ln3   *)
    mutable t : arbre list (* tranche d'paisseur ln2 *)
  } 

  (* retourne le kme nombre de Hamming *)
  let hamming k =
    
  let t = {m2=zro; m5=ln5; n=Int(0); p=Int(0); l=singleton(zro); t=[]} in
  while t.n </ k do

    let m2 = t.m2 in
    t.m2 <- t.m2 ++ ln2;
    let (g,d) = dcoupe t.m2 t.l in
    t.t <- dec m2 g :: t.t;
    t.l <- recolle d (inc ln3 g);
    t.p <- t.p +/ taille(g);
    t.n <- t.n +/ t.p;
    if t.m5 << t.m2++ln3 then (t.l <- insre t.m5 t.l; t.m5 <- t.m5 ++ ln5)

  done;
                      
  if t.n =/ k then t.m2
  else let a = kme_fort (t.n-/k-/Int(1)) t.p t.t in a ++ t.m2 -- ln2

end (* Hamming *)

                    (* +----------------------+
                       |  entiers de Hamming  |
                       +----------------------+ *)

module Hnum : Groupe = struct

  type t = int * int * int (* exposants de 2 3 5 *)
  let zro = (0, 0, 0)
  and ln2  = (1, 0, 0)
  and ln3  = (0, 1, 0)
  and ln5  = (0, 0, 1)
  and (++) (a,b,c) (x,y,z) = (a+x,b+y,c+z)
  and (--) (a,b,c) (x,y,z) = (a-x,b-y,c-z)
  and string(a,b,c) = Printf.sprintf "2^%d . 3^%d . 5^%d" a b c

  (* approximations par dfaut  14 dcimales des logarithmes en base 2 *)
  let lg2 = num_of_string "100000000000000"
  and lg3 = num_of_string "158496250072115"
  and lg5 = num_of_string "232192809488736"
      
  (* comparaison approche puis exacte en cas de doute *)
  let (<<) (a,b,c) (x,y,z) =
    let a = a-x and b = b-y and c = c-z in
    let m = abs(b) + abs(c)
    and l = lg2*/Int(a) +/ lg3*/Int(b) +/ lg5*/Int(c) in
    if      l </ Int(-m) then true
    else if l >/ Int(m)  then false
    else begin
      Printf.fprintf stderr "prcision insuffisante a=%d b=%d c=%d\n" a b c;
      Int(2)**/Int(a) */ Int(3)**/Int(b) */ Int(5)**/Int(c) </ Int(1)
    end

end (* Hnum *)


                     (* +---------------------+
                        |  Lance les calculs  |
                        +---------------------+ *)

(* conversion string -> num avec suffixe *)
let num_of_string(s) =
  let l = String.length(s) in
  let (e,i) = match s.[l-1] with
    |('0'..'9') -> 0, l
    | 'K' -> 3,  l-1
    | 'M' -> 6,  l-1
    | 'G' -> 9,  l-1
    | 'T' -> 12, l-1
    | 'P' -> 15, l-1
    | 'E' -> 18, l-1
    | 'Z' -> 21, l-1
    | 'Y' -> 24, l-1
    |_    -> failwith "suffixe inconnu"
  in Num.num_of_string(String.sub s 0 i) */ (Int 10)**/(Int e)

module H = Hamming(Hnum)
let _ =
  let k = num_of_string(Sys.argv.(1)) in
  let a = H.hamming(k) in
  print_endline(Hnum.string a)
